"""Class models used by Safepull."""
from dataclasses import dataclass
from io import BytesIO

import requests
from rich.table import Table


@dataclass
class Distribution:
    """Dataclass containing distribution information for a PyPI package."""

    filename: str
    packagetype: str
    url: str
    size: int

    @classmethod
    def from_dict(cls, releases: dict):
        """Create a distribution from JSON."""
        return cls(
            filename=releases["filename"],
            packagetype=releases["packagetype"],
            url=releases["url"],
            size=releases["size"],
        )

    def download_package(self) -> tuple[BytesIO, str]:
        """Download a compressed package."""
        r = BytesIO(requests.get(self.url, stream=True, timeout=60).content)
        return r, self.filename

    def get_metadata(self) -> tuple[str, str, str, str]:
        """Return the metadata for a specific distribution of a package."""
        return (
            f"Filename: {self.filename}",
            f"Package Type: {self.packagetype}",
            f"URL: {self.url}",
            f"Size: {self.size / (1 << 20):,.0f}MB",
        )


@dataclass
class Package:
    """Dataclass containing PyPI package metadata."""

    name: str
    summary: str
    author: str
    version: str
    distributions: list[Distribution]

    @classmethod
    def from_dict(cls, package_dict: dict):
        """Create a Package class from JSON."""
        return cls(
            name=package_dict["info"]["name"],
            summary=package_dict["info"]["summary"],
            author=package_dict["info"]["author"],
            version=package_dict["info"]["version"],
            distributions=[
                Distribution.from_dict(releases) for releases in package_dict["urls"]
            ],
        )

    def get_metadata(self) -> tuple[str, str, str]:
        """Return pertinent metadata for the package."""
        return f"{self.name} {self.version}", self.summary, f"Author: {self.author}"

    def get_distributions(self) -> list[Distribution]:
        """Return a list of distributions."""
        return self.distributions

    def get_sdist(self) -> Distribution | None:
        """Extract the source distribution if it exists."""
        for distro in self.distributions:
            if distro.packagetype == "sdist":
                return distro
        return None

    def table_print(self) -> Table:
        """Rich table printing for distribution information."""
        table = Table(
            title=f"{self.name} v.{self.version}\nAuthor: {self.author}\n{self.summary}",
        )  # ruff: noqa: E501
        table.add_column("Index")
        table.add_column("Filename")
        table.add_column("Package Type")
        table.add_column("Size (MB)")
        for ind, entries in enumerate(self.distributions):
            table.add_row(
                str(ind),
                entries.filename,
                entries.packagetype,
                f"{entries.size / (1 << 20):,.2f}",
            )
        return table



"""Safepull module for the safe handling of compressed Python scripts."""
# ruff: noqa: T201

import argparse
import tarfile
from io import BytesIO
from zipfile import ZipFile

import requests
from rich.console import Console

from .models import Package

HOST = "https://pypi.org"


def query_package(package_title: str, version: str | None = None) -> Package:
    """Query PyPI for a package."""
    if version:
        response = requests.get(
            f"{HOST}/pypi/{package_title}/{version}/json",
            timeout=60,
        )
    else:
        response = requests.get(f"{HOST}/pypi/{package_title}/json", timeout=60)
    response.raise_for_status()
    return Package.from_dict(response.json())


def unpack(byte_object: BytesIO, filename: str) -> None:
    """Unpack a compressed file into the CWD."""
    if filename.endswith(".tar.gz"):
        with tarfile.open(fileobj=byte_object) as sdist_tar:
            if hasattr(tarfile, "data_filter"):
                sdist_tar.extractall(filter="data")
            else:
                print(
                    "You do not have data filters enabled for your Python version.",
                    "This is a security risk. Please update Python.",
                    "For more information: https://directorytraversal.zip",
                    sep="\n",
                )
                if input('To ignore this, enter "Y": ').upper() == "Y":
                    sdist_tar.extractall()
    if filename.endswith((".whl", ".zip")):
        with ZipFile(byte_object) as whl_zip:
            whl_zip.extractall()


def run() -> None:
    """Run the program."""
    console = Console()
    parser = argparse.ArgumentParser(
        prog="Safepull",
        description="Extracts a package to the CWD without interfacing with setup.py",
    )
    parser.add_argument("package", help="Package title to be downloaded.")
    parser.add_argument(
        "-v",
        "--version",
        help="Version of a package to be downloaded.",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Automatically selects a download URL.",
    )
    parser.add_argument(
        "-m",
        "--metadata",
        action="store_true",
        help="Displays metadata on a package.",
    )
    args = parser.parse_args()

    use_in = args.package
    if not args.package:
        use_in = input("Input a package title: ")

    user_package = query_package(use_in, args.version)
    if args.metadata:
        print(*user_package.get_metadata(), sep="\n")
        return
    if not args.force:
        if len(distribution_list := user_package.get_distributions()) > 1:
            console.print(user_package.table_print())
            while True:
                use_select = int(input("Enter the index of a package to download: "))
                byteobject, file_name = distribution_list[use_select].download_package()
                unpack(byteobject, file_name)
                break
        else:
            distros = user_package.get_distributions()
            print(*distros[0].get_metadata())
            if input("Download package? (Y/N):").lower() == "y":
                byteobject, file_name = distros[0].download_package()
                unpack(byteobject, file_name)
    else:
        sdist = user_package.get_sdist()
        if sdist is None:
            print("No sdist found. Grabbing first package.")
            sdist = user_package.get_distributions()[0]
        byteobject, file_name = sdist.download_package()
        unpack(byteobject, file_name)
